"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const ajv_1 = __importDefault(require("ajv"));
const json_schema_1 = require("./json-schema");
const tsjson_parser_1 = require("./tsjson-parser");
const SAMPLE_STRING_1 = "Baseball is ninety percent mental and the other half is physical.";
// Sanity-check to make sure the type compatible with what we expect.
// should eventually use something stricter like `tsd` but I'd probably have to roll my own to
// get it to work for this use case.
// eslint-disable-next-line @typescript-eslint/no-unused-vars
const expectType = (_) => { };
const ajv = new ajv_1.default();
describe("Sanity-checks:", () => {
    test("Parse and serialize with type", () => {
        const toSerialize = { something: 1, somethingElse: "2" };
        const serialized = tsjson_parser_1.TSJSON.stringify(toSerialize); // serialized: TsjsonString<typeof toSerialize>
        const parsed = tsjson_parser_1.TSJSON.parse(serialized); // parsed: typeof toSerialize
        expectType(parsed);
        expect(parsed.something).toBe(1);
        expect(parsed.somethingElse).toBe("2");
    });
    test("A string is a valid schema", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({ type: "string" }));
        const parsed = parser.parse(JSON.stringify(SAMPLE_STRING_1));
        expectType(parsed);
        expect(parsed).toBe(SAMPLE_STRING_1);
        expect(parser.schema).toMatchObject({ type: "string" });
        expect(ajv.validateSchema(parser.schema)).toBe(true);
    });
    test("Enums convert to union types", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({ type: "string", enum: ["a", "b", "c"] }));
        expect(() => parser.parse(JSON.stringify(SAMPLE_STRING_1))).toThrow();
        const parsed = parser.parse(JSON.stringify("a"));
        expectType(parsed);
        expect(parsed).toBe("a");
        expect(parser.schema).toMatchObject({
            type: "string",
            enum: ["a", "b", "c"]
        });
        expect(ajv.validateSchema(parser.schema)).toBe(true);
    });
    test("A number is a valid schema", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({ type: "number" }));
        const parsed = parser.parse(JSON.stringify(42));
        expectType(parsed);
        expect(parsed).toBe(42);
        expect(parser.schema).toMatchObject({ type: "number" });
        expect(ajv.validateSchema(parser.schema)).toBe(true);
    });
    test("An integer is a valid schema", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({ type: "integer" }));
        const parsed = parser.parse(JSON.stringify(42));
        expectType(parsed);
        expect(parsed).toBe(42);
        expect(parser.schema).toMatchObject({ type: "integer" });
        expect(ajv.validateSchema(parser.schema)).toBe(true);
    });
    test("Null is a valid schema", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({ type: "null" }));
        const parsed = parser.parse(JSON.stringify(null));
        expectType(parsed);
        expect(parsed).toBe(null);
        expect(parser.schema).toMatchObject({ type: "null" });
        expect(ajv.validateSchema(parser.schema)).toBe(true);
    });
    test("A boolean is a valid schema", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({ type: "boolean" }));
        const parsed = parser.parse(JSON.stringify(true));
        expect(parsed).toBe(true);
        expectType(parsed);
        expect(parser.schema).toMatchObject({ type: "boolean" });
        expect(ajv.validateSchema(parser.schema)).toBe(true);
    });
    test("Validation that doesn't appear in types is still respected", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({ type: "number", minimum: 3 }));
        expect(() => parser.parse(JSON.stringify(2))).toThrow();
        const parsed = parser.parse(JSON.stringify(3));
        expectType(parsed);
        expect(parsed).toBe(3);
        expect(ajv.validateSchema(parser.schema)).toBe(true);
    });
    test("An empty array is a valid schema", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({ type: "array" }));
        const parsed = parser.parse(JSON.stringify([]));
        expect(parsed).toStrictEqual([]);
        expectType(parsed);
        expect(parser.schema).toMatchObject({
            type: "array"
        });
        expect(ajv.validateSchema(parser.schema)).toBe(true);
    });
    test("An object is a valid schema", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({ type: "object", required: [] }));
        const parsed = parser.parse(JSON.stringify({}));
        expect(parsed).toStrictEqual({});
        expectType(parsed);
        expect(parser.schema).toMatchObject({ type: "object" });
        expect(ajv.validateSchema(parser.schema)).toBe(true);
    });
    test("AnyOf works with simple schemas", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({
            anyOf: [json_schema_1.createSchema({ type: "string" }), json_schema_1.createSchema({ type: "number" })]
        }));
        let parsed = parser.parse(JSON.stringify(SAMPLE_STRING_1));
        expectType(parsed);
        expect(parsed).toBe(SAMPLE_STRING_1);
        parsed = parser.parse(JSON.stringify(42));
        expect(parsed).toBe(42);
        expect(() => parser.parse(JSON.stringify({}))).toThrow();
        expect(ajv.validateSchema(parser.schema)).toBe(true);
    });
    test("AllOf works with simple schemas", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({
            allOf: [
                json_schema_1.createSchema({
                    type: "object",
                    properties: {
                        a: json_schema_1.createSchema({ type: "string", enum: ["a", "b", "c"] }),
                        b: json_schema_1.createSchema({ type: "number" }),
                        d: json_schema_1.createSchema({ type: "number" })
                    },
                    required: []
                }),
                json_schema_1.createSchema({
                    type: "object",
                    properties: {
                        a: json_schema_1.createSchema({ type: "string" }),
                        c: json_schema_1.createSchema({ type: "string" }),
                        d: json_schema_1.createSchema({ type: "string" })
                    },
                    required: ["a"]
                })
            ]
        }));
        const parsed = parser.parse(JSON.stringify({ a: "a" }));
        expectType(parsed);
        expect(parsed).toStrictEqual({ a: "a" });
        expect(() => parser.parse(JSON.stringify({}))).toThrow();
        expect(ajv.validateSchema(parser.schema)).toBe(true);
    });
    test("Ensure validate is not called if skipValidation is true", () => {
        /* eslint-disable @typescript-eslint/unbound-method */
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({ type: "null" }));
        parser.validate = jest.fn();
        const parsed = parser.parse(JSON.stringify(null));
        expectType(parsed);
        expect(parsed).toBe(null);
        expect(parser.validate).toBeCalledTimes(1);
        parser.parse(JSON.stringify(null), true);
        expect(parser.validate).toBeCalledTimes(1);
        expect(ajv.validateSchema(parser.schema)).toBe(true);
        /* eslint-enable @typescript-eslint/unbound-method */
    });
    test("SkipValidation is dangerous", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({ type: "null" }));
        const parsed = parser.parse(JSON.stringify(SAMPLE_STRING_1), true);
        expectType(parsed); // typeof parsed === null, which is wrong!
        expect(parsed).toBe(SAMPLE_STRING_1);
        expect(() => parser.validate(parsed)).toThrow();
        expect(ajv.validateSchema(parser.schema)).toBe(true);
    });
    test("Invalid schema fails validation", () => {
        expect(ajv.validateSchema({ type: "invalid" })).toBe(false);
    });
});
describe("More involved tests with arrays", () => {
    test("Array of objects", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({
            type: "array",
            items: json_schema_1.createSchema({ type: "object", properties: { a: json_schema_1.createSchema({ type: "string" }) } })
        }));
        const toParse = [{ a: "1" }, { a: "2" }, {}];
        const parsed = parser.parse(JSON.stringify(toParse));
        expectType(parsed);
        expect(parsed).toStrictEqual(toParse);
    });
    test("Array of required objects", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({
            type: "array",
            items: json_schema_1.createSchema({
                type: "object",
                properties: { a: json_schema_1.createSchema({ type: "string" }) },
                required: ["a"]
            })
        }));
        const toParseInvalid = [{ a: "1" }, { a: "2" }, {}];
        expect(() => parser.parse(JSON.stringify(toParseInvalid))).toThrow();
        const toParseValid = [{ a: "1" }, { a: "2" }];
        const parsed = parser.parse(JSON.stringify(toParseValid));
        expectType(parsed);
        expect(parsed).toStrictEqual(toParseValid);
    });
    test("Arrays of arrays", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({
            type: "array",
            items: json_schema_1.createSchema({
                type: "array",
                items: json_schema_1.createSchema({ type: "array", items: json_schema_1.createSchema({ type: "number" }) })
            })
        }));
        const toParse = [[[0]], [], [[1]], [[1], [2], []]];
        const parsed = parser.parse(JSON.stringify(toParse));
        expectType(parsed);
        expect(parsed).toStrictEqual(toParse);
        const toFail = [[[0]], [], [[1]], [[1], [2], 0]];
        expect(() => parser.parse(JSON.stringify(toFail))).toThrow();
    });
    test("Array with additionalItems", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({
            type: "array",
            items: [json_schema_1.createSchema({ type: "string" })],
            additionalItems: json_schema_1.createSchema({ type: "number" })
        }));
        const toParse = ["1", 2, 3, 4];
        const parsed = parser.parse(JSON.stringify(toParse));
        expectType(parsed); // this type should really be [string, ...number]
        expect(parsed).toStrictEqual(toParse);
        const toFail = [1, "2"]; // this will pass typechecking even though it will fail validation!
        expect(() => parser.parse(JSON.stringify(toFail))).toThrow();
    });
});
describe("More involved tests with objects", () => {
    test("Object with various fields", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({
            type: "object",
            properties: {
                a: json_schema_1.createSchema({ type: "array" }),
                b: json_schema_1.createSchema({ type: "array" }),
                c: json_schema_1.createSchema({ type: "object" }),
                d: json_schema_1.createSchema({ type: "object" }),
                e: json_schema_1.createSchema({ type: "string" }),
                f: json_schema_1.createSchema({ type: "number" })
            },
            additionalProperties: json_schema_1.createSchema({ type: "number" }),
            required: ["a", "c", "e"]
        }));
        const toParse = {
            a: [],
            b: [1, "a", 4],
            c: { hello: 123 },
            e: "thisisastring"
        };
        const parsed = parser.parse(JSON.stringify(toParse));
        expectType(parsed);
        expect(parsed).toStrictEqual(toParse);
        expectType(parsed.anotherProp);
        const toFail = {};
        expect(() => parser.parse(JSON.stringify(toFail))).toThrow();
    });
});
describe("Odd combinations of things", () => {
    test("Enum without type", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({ enum: [1, "2", { x: "y" }] }));
        const toParse = { x: "y" };
        const parsed = parser.parse(JSON.stringify(toParse));
        expectType(parsed);
        expect(parsed).toStrictEqual(toParse);
    });
    test("Const object", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({ const: { a: 3 } }));
        const toParse = { a: 3 };
        const parsed = parser.parse(JSON.stringify(toParse));
        expectType(parsed);
        expect(parsed).toStrictEqual(toParse);
        const toFail = { a: 4 };
        expect(() => parser.parse(JSON.stringify(toFail))).toThrow();
    });
    test("Const array", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({ const: ["a", 1] }));
        const toParse = ["a", 1];
        const parsed = parser.parse(JSON.stringify(toParse));
        expectType(parsed);
        expect(parsed).toStrictEqual(toParse);
        const toFail = ["a"];
        expect(() => parser.parse(JSON.stringify(toFail))).toThrow();
    });
    test("AllOf and anyOf", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({
            anyOf: [
                json_schema_1.createSchema({ type: "string", enum: ["a", "b", "c"] }),
                json_schema_1.createSchema({ type: "string", enum: ["a", "b", "d"] })
            ],
            // allOf: [S({ type: "string", enum: ["a", "d"] as const })] // is currently deriving a never type, which is wrong
            allOf: [json_schema_1.createSchema({ type: "string" })]
        }));
        const parsed = parser.parse(JSON.stringify("a"));
        expectType(parsed);
        expect(parsed).toBe("a");
    });
    test("Subschemas without type inherit type", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({
            type: "string",
            anyOf: [
                json_schema_1.createSchema({ enum: ["a"] }),
                json_schema_1.createSchema({ description: "unconstrained" })
            ]
        }));
        const parsed = parser.parse(JSON.stringify("z"));
        expectType(parsed);
        expect(parsed).toBe("z");
    });
});
describe("Ref tests", () => {
    // Hope to be able to delete this test soon, and also maybe enforce that every ref
    // points to a depencency (but TBD whether this is actually desired, since refs can be valid outside of the schema)
    // Example taken from https://json-schema.org/understanding-json-schema/structuring.html
    test("Any ref is unknown :( ", () => {
        const parser = new tsjson_parser_1.TsjsonParser(json_schema_1.createSchema({
            $schema: "http://json-schema.org/draft-07/schema#",
            definitions: {
                person: json_schema_1.createSchema({
                    type: "object",
                    properties: {
                        name: json_schema_1.createSchema({ type: "string" }),
                        children: json_schema_1.createSchema({
                            type: "array",
                            items: json_schema_1.createSchema({ $ref: "#/definitions/person" }),
                            default: []
                        })
                    }
                })
            },
            type: "object",
            properties: {
                person: json_schema_1.createSchema({ $ref: "#/definitions/person" })
            }
        }));
        const toParse = {
            person: {
                name: "Elizabeth",
                children: [
                    {
                        name: "Charles",
                        children: [
                            {
                                name: "William",
                                children: [{ name: "George" }, { name: "Charlotte" }]
                            },
                            {
                                name: "Harry"
                            }
                        ]
                    }
                ]
            }
        };
        const parsed = parser.parse(JSON.stringify(toParse));
        expectType(parsed);
        expect(parsed).toStrictEqual(toParse);
    });
});
